/***************************************************************************
 *   Copyright (C) 2017 by ON Semiconductor                                *
 *   fg64rh@onsemi.com                                                     *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "imp.h"
#include <helper/binarybuffer.h>
#include <target/algorithm.h>
#include <target/armv7m.h> /* TODO armv6m? */
#include <target/cortex_m.h>

/* Flash Macros */
#define AXM0_FLASH_SZ					(uint16_t)64		/* 64 KB */
#define AXM0_RAM_SZ						(uint16_t)8			/* 8 KB */
#define AXM0_PART_NUM					(uint16_t)0xC60
#define AXM0_PART_FAMILY				(uint16_t)0x0		/* CORTEX M0 */
#define AXM0_PART_REV					(uint8_t)0x1		/* Hardware revision */
#define AXM0_FLASH_BANK_START_ADDRESS	(uint32_t)0x00000000	/* Flash bank start address */
#define AXM0_FLASH_BANK_END_ADDRESS		(uint32_t)(0x0000FFFF - 4) /* !!! Temporary "-4" as changing content @0xFFFC locks debug interface */
#define AXM0_FLASH_INFO_BLK_START_ADRS	(uint32_t)0x08000000	/* Flash info block start address */
#define AXM0_FLASH_INFO_BLK_END_ADRS	(uint32_t)0x080001FF

#define AXM0_1K_BYTES					(int)1024
#define AXM0_PERIPHERAL_FLASH_BIT_POS	(uint32_t)(1 << 12)

/* Register memory map */
#define AXM0_PERIPHERAL_CLOCK			(uint32_t)0x40300010
#define AXM0_FLASH_CONTROL				(uint32_t)0x40C00000
#define AXM0_FLASH_ADDRESS				(uint32_t)0x40C00004
#define AXM0_FLASH_DATA					(uint32_t)0x40C00008
#define AXM0_FLASH_PAGE_PROTECT			(uint32_t)0x40C00018
#define AXM0_FLASH_CONTROL_KEYWORD		(uint32_t)0xA45B0000
#define AXM0_FLASH_CMD_PAGE_ERASE		(uint32_t)(1 << 1)
#define AXM0_FLASH_CMD_MASS_ERASE		(uint32_t)(1 << 2)
#define AXM0_FLASH_PROG_ERASE_DONE	(uint32_t)(1 << 5)
#define AXM0_PAGE_PROTECT_ALL			(uint32_t)0xFFFFFFFF
#define AXM0_PAGE_PROTECT_CLR_ALL		(uint32_t)0x00000000
#define AXM0_FLASH_CONTROL_CONF			(uint32_t)0xA45B0000
#define AXM0_FLASH_CONTROL_PROG			(uint32_t)0xA45B0001
#define AXM0_FLASH_CONTROL_PROG_16BIT	(uint32_t)0xA45B0009
#define AXM0_FLASH_BASE                  0x40C00000UL

#define AXM0_NUM_OF_PAGE_PROTECT_BIT_IN_REG	(uint8_t)32
#define AXM0_WORD_SZ						(uint8_t)4		/* 4 bytes */
#define AXM0_FLASH_BANK						(uint8_t)0
#define AXM0_FLASH_INFO_BLOCK				(uint8_t)1
#define AXM0_DECODE_BUF_SIZE				(int)256

/* Flash page macros */
#define AXM0_FLASH_PAGE_SIZE				(uint16_t)0x200

struct axm0_flash_bank {
	int probed;
};

struct axm0_info {
	uint16_t flash_sz_kib;
	uint16_t ram_sz_kib;
	uint16_t part_num;
	uint16_t part_family;
	uint8_t prod_rev;
	uint16_t page_size;
};

static int axm0_write(struct flash_bank *bank, const uint8_t *buffer,
	uint32_t offset, uint32_t count);

static int axm0_get_flash_size(struct flash_bank *bank, uint16_t *flash_sz)
{
	*flash_sz = AXM0_FLASH_SZ;
	return ERROR_OK;
}

static int axm0_get_ram_size(struct flash_bank *bank, uint16_t *ram_sz)
{
	*ram_sz = AXM0_RAM_SZ;
	return ERROR_OK;
}

static int axm0_get_part_num(struct flash_bank *bank, uint16_t *pnum)
{
	*pnum = AXM0_PART_NUM;
	return ERROR_OK;
}

static int axm0_get_part_family(struct flash_bank *bank, uint16_t *pfamily)
{
	*pfamily = AXM0_PART_FAMILY;
	return ERROR_OK;
}

static int axm0_get_prod_rev(struct flash_bank *bank, uint8_t *prev)
{
	*prev = AXM0_PART_REV;
	return ERROR_OK;
}

static int axm0_read_info(struct flash_bank *bank,
	struct axm0_info *axm0_info)
{
	int ret;

	memset(axm0_info, 0, sizeof(struct axm0_info));

	ret = axm0_get_flash_size(bank, &(axm0_info->flash_sz_kib));
	if (ERROR_OK != ret)
		return ret;

	ret = axm0_get_ram_size(bank, &(axm0_info->ram_sz_kib));
	if (ERROR_OK != ret)
		return ret;

	ret = axm0_get_part_num(bank, &(axm0_info->part_num));
	if (ERROR_OK != ret)
		return ret;

	ret = axm0_get_part_family(bank, &(axm0_info->part_family));
	if (ERROR_OK != ret)
		return ret;

	ret = axm0_get_prod_rev(bank, &(axm0_info->prod_rev));
	if (ERROR_OK != ret)
		return ret;

	axm0_info->page_size = AXM0_FLASH_PAGE_SIZE;

	return ERROR_OK;
}

/*
 * Helper to create a human friendly string describing a part
 */

static int axm0_decode_info(struct axm0_info *info, char *buf,
								 int buf_size)
{
	int printed = 0;

	printed = snprintf(buf, buf_size, "axm0 "); /* Part name */

	buf += printed;
	buf_size -= printed;

	if (0 >= buf_size)
	return ERROR_BUF_TOO_SMALL;

	printed = snprintf(buf, buf_size, "Nebo"); /* Part family */

	buf += printed;
	buf_size -= printed;

	if (0 >= buf_size)
		return ERROR_BUF_TOO_SMALL;

	/* Product revision */
	printed = snprintf(buf, buf_size, " - Rev: %d", info->prod_rev);
	buf += printed;
	buf_size -= printed;

	if (0 >= buf_size)
		return ERROR_BUF_TOO_SMALL;

	return ERROR_OK;
}

/* flash bank axm0 <base> <size> 0 0 <target#>
 */
FLASH_BANK_COMMAND_HANDLER(axm0_flash_bank_command)
{
	struct axm0_flash_bank *axm0_info;

	if (CMD_ARGC < 6)
		return ERROR_COMMAND_SYNTAX_ERROR;

	axm0_info = malloc(sizeof(struct axm0_flash_bank));

	bank->driver_priv = axm0_info;
	axm0_info->probed = 0;

	return ERROR_OK;
}

static int axm0_erase(struct flash_bank *bank, int first, int last)
{
	/* 1. Enable flash clock
	   2. Unlock flash
	   3. Write FLASH ADDRESS register
	   4. write COMMAND_PAGE_ERASE
	   5. wait until STATUS_NOT_BUSY
	 */

	int ret, page;
	uint32_t index;
	uint32_t clock_config, flash_status, page_protect[4];

	/* Read peripheral clock register */
	ret = target_read_u32(bank->target, AXM0_PERIPHERAL_CLOCK, &clock_config);
	if (ERROR_OK != ret) {
		LOG_ERROR("Failed to read peripheral clock register");
		return ret;
	}

	/* Enable flash clock */
	ret = target_write_u32(bank->target, AXM0_PERIPHERAL_CLOCK, \
			clock_config | AXM0_PERIPHERAL_FLASH_BIT_POS);
	if (ERROR_OK != ret) {
		LOG_ERROR("Failed to enable flash clock");
		return ret;
	}

	/* Reset page protect registers */
	for(index = 0; index < 3; index++)
	{
		/* Read page protect register */
		ret = target_read_u32(bank->target, AXM0_FLASH_PAGE_PROTECT + \
						(index * 4), (page_protect + index));
		if (ERROR_OK != ret) {
			LOG_ERROR("Failed to read page protect register");
			return ret;
		}

		/* Reset page protect bit */
		ret = target_write_u32(bank->target, AXM0_FLASH_PAGE_PROTECT + \
							(index * 4), AXM0_PAGE_PROTECT_CLR_ALL);
		if (ERROR_OK != ret) {
			LOG_ERROR("Failed to reset page protect bit");
			return ret;
		}
	}

	for(page = first; page <= last; page++)
	{
		/* Set address to be erased */
		ret = target_write_u32(bank->target, AXM0_FLASH_ADDRESS, \
			  (page * AXM0_FLASH_PAGE_SIZE));
		if (ERROR_OK != ret) {
			LOG_ERROR("Failed to write 'flash address' register");
			return ret;
		}

		/* Execute erase command */
		/* special case full flash erases for speed */
		index = AXM0_FLASH_CONTROL_KEYWORD | AXM0_FLASH_CMD_PAGE_ERASE;
		if (page <= 0 && last >= 127)
		{
			index = AXM0_FLASH_CONTROL_KEYWORD | AXM0_FLASH_CMD_MASS_ERASE;
			page = last;
		}

		ret = target_write_u32(bank->target, AXM0_FLASH_CONTROL, index);
		if (ERROR_OK != ret) {
			LOG_ERROR("Failed to erase flash");
			return ret;
		}

		/* Check flash erase is complete */
		do
		{
			ret = target_read_u32(bank->target, AXM0_FLASH_CONTROL, \
									&flash_status);
			if (ERROR_OK != ret) {
				LOG_ERROR("Failed to read flash control");
				return ret;
			}
		}while(!(flash_status & AXM0_FLASH_PROG_ERASE_DONE)); /* Wait till erase is done */
	}

	for(index = 0; index < 3; index++)
	{
		/* Reset page protect bit */
		ret = target_write_u32(bank->target, AXM0_FLASH_PAGE_PROTECT + \
							(index * 4), page_protect[index]);
		if (ERROR_OK != ret) {
			LOG_ERROR("Failed to reset page protect bit");
			return ret;
		}
	}

	/* Restore FLASH clock */
	ret = target_write_u32(bank->target, AXM0_PERIPHERAL_CLOCK, clock_config);
	if (ERROR_OK != ret) {
		LOG_ERROR("Failed to restore flash clock");
		return ret;
	}

	return ERROR_OK;
}

static int axm0_protect(struct flash_bank *bank, int set, int first, int last)
{
	int ret;
	uint32_t index;
	uint32_t clock_config;

	if(!set) {
		LOG_ERROR("Erase/write device data to unlock flash");
		return ERROR_FAIL;
	}

	/* Read peripheral clock register */
	ret = target_read_u32(bank->target, AXM0_PERIPHERAL_CLOCK, &clock_config);
	if (ERROR_OK != ret) {
		LOG_ERROR("Failed to read peripheral clock register");
		return ret;
	}

	/* Enable FLASH clock */
	ret = target_write_u32(bank->target, AXM0_PERIPHERAL_CLOCK, \
						   clock_config | AXM0_PERIPHERAL_FLASH_BIT_POS);
	if (ERROR_OK != ret) {
		LOG_ERROR("Failed to enable flash clock");
		return ret;
	}

	/* Protect flash */
	for(index = 0; index < 3; index++)
	{
		ret = target_write_u32(bank->target, (AXM0_FLASH_PAGE_PROTECT + \
						(index * 4)), AXM0_PAGE_PROTECT_ALL);
		if (ERROR_OK != ret) {
			LOG_ERROR("Failed to protect flash");
			return ret;
		}
	}

	/* Restore FLASH clock */
	ret = target_write_u32(bank->target, AXM0_PERIPHERAL_CLOCK, clock_config);
	if (ERROR_OK != ret) {
		LOG_ERROR("Failed to restore flash clock");
		return ret;
	}

	return ERROR_OK;
}

static int axm0_write_block(struct flash_bank *bank, const uint8_t *buf, uint32_t offset, uint32_t count)
{
	struct target *target = bank->target;
	uint32_t buffer_size = 4096;
	struct working_area *write_algorithm;
	struct working_area *source;
	uint32_t address = bank->base + offset;
	struct reg_param reg_params[5];
	struct armv7m_algorithm armv7m_info;
	int ret = ERROR_OK;

	static const uint8_t axm0_flash_write_code[] = {
		0x16, 0x68, 0x00, 0x2E, 0x18, 0xD0, 0x55, 0x68, 0xB5, 0x42, 0xF9, 0xD0,
		0x44, 0x60, 0x2E, 0x68, 0x86, 0x60, 0x0A, 0x4E, 0x06, 0x60, 0x04, 0x35,
		0x04, 0x34, 0x06, 0x68, 0x20, 0x27, 0x3E, 0x42, 0xFB, 0xD0, 0x9D, 0x42,
		0x01, 0xD3, 0x15, 0x46, 0x08, 0x35, 0x55, 0x60, 0x01, 0x39, 0x00, 0x29,
		0x02, 0xD0, 0xE5, 0xE7, 0x00, 0x20, 0x50, 0x60, 0x30, 0x46, 0x00, 0xBE,
		0x01, 0x00, 0x5B, 0xA4
	};

	/* flash write code */
	if (target_alloc_working_area(target, sizeof(axm0_flash_write_code),
				      &write_algorithm) != ERROR_OK) {
		LOG_WARNING("no working area available, can't do block memory writes");
		return ERROR_TARGET_RESOURCE_NOT_AVAILABLE;
	}

	ret = target_write_buffer(target, write_algorithm->address,
				  sizeof(axm0_flash_write_code), axm0_flash_write_code);
	if (ret != ERROR_OK)
		return ret;

	/* memory buffer */
	while (target_alloc_working_area_try(target, buffer_size, &source) != ERROR_OK) {
		buffer_size /= 2;
		buffer_size &= ~3UL; /* Make sure it's 4 byte aligned */
		if (buffer_size <= 256) {
			/* we already allocated the writing code, but failed to get a
			 * buffer, free the algorithm */
			target_free_working_area(target, write_algorithm);

			LOG_WARNING("no large enough working area available, can't do block memory writes");
			return ERROR_TARGET_RESOURCE_NOT_AVAILABLE;
		}
	}

	init_reg_param(&reg_params[0], "r0", 32, PARAM_IN_OUT);	/* flash base (in), status (out) */
	init_reg_param(&reg_params[1], "r1", 32, PARAM_OUT);	/* count (word-32bit) */
	init_reg_param(&reg_params[2], "r2", 32, PARAM_OUT);	/* buffer start */
	init_reg_param(&reg_params[3], "r3", 32, PARAM_OUT);	/* buffer end */
	init_reg_param(&reg_params[4], "r4", 32, PARAM_IN_OUT);	/* target address */

	buf_set_u32(reg_params[0].value, 0, 32, AXM0_FLASH_BASE);
	buf_set_u32(reg_params[1].value, 0, 32, count);
	buf_set_u32(reg_params[2].value, 0, 32, source->address);
	buf_set_u32(reg_params[3].value, 0, 32, source->address + source->size);
	buf_set_u32(reg_params[4].value, 0, 32, address);

	armv7m_info.common_magic = ARMV7M_COMMON_MAGIC;
	armv7m_info.core_mode = ARM_MODE_THREAD;

	ret = target_run_flash_async_algorithm(target, buf, count, 4,
			0, NULL,
			5, reg_params,
			source->address, source->size,
			write_algorithm->address, 0,
			&armv7m_info);

	if (ret == ERROR_FLASH_OPERATION_FAILED) {
		LOG_ERROR("flash write failed at address 0x%"PRIx32,
				buf_get_u32(reg_params[4].value, 0, 32));
	}

	target_free_working_area(target, source);
	target_free_working_area(target, write_algorithm);

	destroy_reg_param(&reg_params[0]);
	destroy_reg_param(&reg_params[1]);
	destroy_reg_param(&reg_params[2]);
	destroy_reg_param(&reg_params[3]);
	destroy_reg_param(&reg_params[4]);

	return ret;
}

static int axm0_write(struct flash_bank *bank, const uint8_t *buffer,
		uint32_t offset, uint32_t count)
{
	struct target *target = bank->target;
	int ret;
	uint32_t index;
	uint32_t clock_config, page_protect[4], flash_status;
	uint8_t *new_buffer = NULL;

	if (target->state != TARGET_HALTED) {
		LOG_ERROR("Target not halted");
		return ERROR_TARGET_NOT_HALTED;
	}

	if (offset & 0x3) {
		LOG_ERROR("offset 0x%" PRIx32 " breaks required 4-byte "
			"alignment", offset);
		return ERROR_FLASH_DST_BREAKS_ALIGNMENT;
	}

	/* Check flash boundaries *//* TODO need to handle last word in 64K and 0.5K info block */
	if((((offset + count) > AXM0_FLASH_BANK_END_ADDRESS) && \
	   (bank->bank_number == AXM0_FLASH_BANK)) ||
	   ((((offset + count) > AXM0_FLASH_INFO_BLK_END_ADRS) || \
	   (offset < AXM0_FLASH_INFO_BLK_START_ADRS)) && \
	   (bank->bank_number == AXM0_FLASH_INFO_BLOCK))) {
		/* Outside flash boundary */
		LOG_ERROR("Trying to write outside the flash for bank number %d at address 0x%08X", \
					bank->bank_number, (offset + count));
		return ERROR_FAIL;
	}

	/* Read peripheral clock register */
	ret = target_read_u32(bank->target, AXM0_PERIPHERAL_CLOCK, &clock_config);
	if (ERROR_OK != ret) {
		LOG_ERROR("Failed to read peripheral clock register");
		return ret;
	}

	/* Enable FLASH clock */
	ret = target_write_u32(bank->target, AXM0_PERIPHERAL_CLOCK, \
					clock_config | AXM0_PERIPHERAL_FLASH_BIT_POS);
	if (ERROR_OK != ret) {
		LOG_ERROR("Failed to enable flash clock");
		return ret;
	}

	/* Reset page protect registers */
	for(index = 0; index < 3; index++)
	{
		/* Read page protect register */
		ret = target_read_u32(bank->target, AXM0_FLASH_PAGE_PROTECT + \
							(index * AXM0_WORD_SZ), (page_protect + index));
		if (ERROR_OK != ret) {
			LOG_ERROR("Failed to read page protect register");
			return ret;
		}

		/* Reset page protect bit */
		ret = target_write_u32(bank->target, AXM0_FLASH_PAGE_PROTECT + \
					(index * AXM0_WORD_SZ), AXM0_PAGE_PROTECT_CLR_ALL);
		if (ERROR_OK != ret) {
			LOG_ERROR("Failed to reset page protect bit");
			return ret;
		}
	}

	if (count & 0x3) {
		uint32_t old_count = count;
		count = (old_count | 3) + 1;
		new_buffer = malloc(count);
		if (new_buffer == NULL) {
			LOG_ERROR("odd number of bytes to write and no memory "
				"for padding buffer");
			return ERROR_FAIL;
		}
		LOG_INFO("odd number of bytes to write (%" PRIu32 "), extending to %" PRIu32 " "
			"and padding with 0xff", old_count, count);
		memset(new_buffer, 0xff, count);
		buffer = memcpy(new_buffer, buffer, old_count);
	}

	uint32_t words_remaining = count / 4;

	/* try using a block write */
	ret = axm0_write_block(bank, buffer, offset, words_remaining);

	if (ret == ERROR_TARGET_RESOURCE_NOT_AVAILABLE) {
		/* Set flash address */
		ret = target_write_u32(bank->target, AXM0_FLASH_ADDRESS, offset);
		if (ERROR_OK != ret) {
			LOG_ERROR("Failed to write address");
			goto cleanup;
		}
		/* Write code into flash: Write multiple of words */
		for (; words_remaining > 0; --words_remaining, buffer += 4) {
			uint32_t value;
			memcpy(&value, buffer, sizeof(uint32_t));
			/* Write multiple of words */
			ret = target_write_u32(bank->target, AXM0_FLASH_DATA, value);
			if (ERROR_OK != ret) {
				LOG_ERROR("Failed to write to flash data");
				goto cleanup;
			}
			/* Write to flash */
			ret = target_write_u32(bank->target, AXM0_FLASH_CONTROL, \
					       AXM0_FLASH_CONTROL_PROG);
			if (ERROR_OK != ret) {
				LOG_ERROR("Failed to write to flash");
				goto cleanup;
			}

			/* Check flash program is complete */
			do {
				ret = target_read_u32(bank->target, AXM0_FLASH_CONTROL, \
						      &flash_status);
				if (ERROR_OK != ret) {
					LOG_ERROR("Failed to read flash control");
					goto cleanup;
				}
			} while(!(flash_status & AXM0_FLASH_PROG_ERASE_DONE)); /* Wait till program is done */
		}
	}

	/* Restore page protect registers */
	for(index = 0; index < 3; index++)
	{
		ret = target_write_u32(bank->target, AXM0_FLASH_PAGE_PROTECT + \
					(index * AXM0_WORD_SZ), page_protect[index]);
		if (ERROR_OK != ret) {
			LOG_ERROR("Failed to reset page protect register");
			goto cleanup;
		}
	}

	/* Restore flash clock */
	ret = target_write_u32(bank->target, AXM0_PERIPHERAL_CLOCK, clock_config);
	if (ERROR_OK != ret) {
		LOG_ERROR("Failed to restore flash clock");
		goto cleanup;
	}


cleanup:
	if (new_buffer)
		free(new_buffer);

	return ERROR_OK;
}

static int axm0_probe(struct flash_bank *bank)
{
	struct axm0_flash_bank *axm0_info = bank->driver_priv;
	struct axm0_info axm0_mcu_info;
	int ret, i;
	int num_pages;
	uint8_t index;
	uint32_t base_address, page_protect[4];
	char buf[AXM0_DECODE_BUF_SIZE];

	axm0_info->probed = 0;

	if(bank->bank_number == AXM0_FLASH_BANK)
	{
		base_address = AXM0_FLASH_BANK_START_ADDRESS;
	}
	else if(bank->bank_number == AXM0_FLASH_INFO_BLOCK)
	{
		base_address = AXM0_FLASH_INFO_BLK_START_ADRS;
	}
	else
	{
		/* Invalid bank number('else' added to satisfy the compiler) */
		LOG_ERROR("Invalid bank number %x", bank->bank_number);
		return ERROR_FAIL;
	}

	ret = axm0_read_info(bank, &axm0_mcu_info);
	if (ERROR_OK != ret)
		return ret;


	ret = axm0_decode_info(&axm0_mcu_info, buf, sizeof(buf));
	if (ERROR_OK != ret)
		return ret;

	LOG_INFO("detected part: %s", buf);
	LOG_INFO("flash size = %dkbytes", axm0_mcu_info.flash_sz_kib);
	LOG_INFO("flash page size = %dbytes", axm0_mcu_info.page_size);

	assert(0 != axm0_mcu_info.page_size);

	num_pages = axm0_mcu_info.flash_sz_kib * AXM0_1K_BYTES / \
				axm0_mcu_info.page_size;

	assert(num_pages > 0);

	if (bank->sectors) {
		free(bank->sectors);
		bank->sectors = NULL;
	}

	bank->base = base_address;
	bank->size = (num_pages * axm0_mcu_info.page_size);
	bank->num_sectors = num_pages;

	bank->sectors = malloc(sizeof(struct flash_sector) * num_pages);

	/* Reset page protect registers */
	for(index = 0; index < 3; index++)
	{
		/* Read page protect register */
		ret = target_read_u32(bank->target, AXM0_FLASH_PAGE_PROTECT + \
						(index * AXM0_WORD_SZ), (page_protect + index));
		if (ERROR_OK != ret) {
			LOG_ERROR("Failed to read page protect register");
			return ret;
		}
	}

	for (i = 0; i < num_pages; i++) {
		bank->sectors[i].offset = i * axm0_mcu_info.page_size;
		bank->sectors[i].size = axm0_mcu_info.page_size;
		bank->sectors[i].is_erased = -1;
		bank->sectors[i].is_protected = (page_protect[i / \
				AXM0_NUM_OF_PAGE_PROTECT_BIT_IN_REG] >> (i % 32)) & 0x1;
	}

	axm0_info->probed = 1;

	return ERROR_OK;
}

static int axm0_auto_probe(struct flash_bank *bank)
{
	struct axm0_flash_bank *axm0_info = bank->driver_priv;

	if (axm0_info->probed)
		return ERROR_OK;

	return axm0_probe(bank);
}

static int axm0_protect_check(struct flash_bank *bank)
{
	int i, ret;
	uint8_t index;
	uint32_t page_protect[4];

	/* Reset page protect registers */
	for(index = 0; index < 3; index++)
	{
		/* Read page protect register */
		ret = target_read_u32(bank->target, AXM0_FLASH_PAGE_PROTECT + \
						(index * AXM0_WORD_SZ), (page_protect + index));
		if (ERROR_OK != ret) {
			LOG_ERROR("Failed to read page protect register");
			return ret;
		}
	}
	/* Update global lock status */
	for (i = 0; i < bank->num_sectors; i++) {
		bank->sectors[i].is_protected = (page_protect[i / \
			AXM0_NUM_OF_PAGE_PROTECT_BIT_IN_REG] >> (i % 32)) & 0x1;
	}

	return ERROR_OK;
}

static int axm0_get_info(struct flash_bank *bank, char *buf, int buf_size)
{
	struct axm0_info info;
	int ret = 0;

	ret = axm0_read_info(bank, &info);
	if (ERROR_OK != ret) {
		LOG_ERROR("Failed to read AXM0 info");
		return ret;
	}

	return axm0_decode_info(&info, buf, buf_size);
}

COMMAND_HANDLER(axm0_handle_debuglock_command)
{
	return ERROR_OK;
}

static const struct command_registration axm0_exec_command_handlers[] = {
	{
		.name = "debuglock",
		.handler = axm0_handle_debuglock_command,
		.mode = COMMAND_EXEC,
		.usage = "bank_id",
		.help = "Lock the debug interface of the device.",
	},
	COMMAND_REGISTRATION_DONE
};

static const struct command_registration axm0_command_handlers[] = {
	{
		.name = "axm0",
		.mode = COMMAND_ANY,
		.help = "axm0 flash command group",
		.usage = "",
		.chain = axm0_exec_command_handlers,
	},
	COMMAND_REGISTRATION_DONE
};

struct flash_driver axm0_flash = {
	.name = "axm0",
	.commands = axm0_command_handlers,
	.flash_bank_command = axm0_flash_bank_command,
	.erase = axm0_erase,
	.protect = axm0_protect,
	.write = axm0_write,
	.read = default_flash_read,
	.probe = axm0_probe,
	.auto_probe = axm0_auto_probe,
	.erase_check = default_flash_blank_check,
	.protect_check = axm0_protect_check,
	.info = axm0_get_info,
};
