/***************************************************************************
 *   Copyright (C) 2017 by ON Semiconductor                                *
 *   fg64rh@onsemi.com                                                     *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "imp.h"
#include <helper/binarybuffer.h>
#include <target/algorithm.h>
#include <target/armv7m.h>
#include <target/cortex_m.h>

/* Flash Macros */
#define NCS36510_FLASH_SZ					(uint16_t)328		/* 328 KB */
#define NCS36510_RAM_SZ						(uint16_t)48		/* 48 KB */
#define NCS36510_PART_NUM					(uint16_t)36510		/* NCS36510 */
#define NCS36510_PART_FAMILY				(uint16_t)36510		/* "NCS" */
#define NCS36510_PART_REV					(uint8_t)0x7		/* Hardware revision 'G' */
#define NCS36510_FLASH_A_START_ADDRESS		(uint32_t)0x0		/* Flash A start address */
#define NCS36510_FLASH_B_START_ADDRESS		(uint32_t)0x100000	/* Flash A start address */
#define NCS36510_FLASH_A_END_ADDRESS		(uint32_t)0x52000
#define NCS36510_FLASH_B_END_ADDRESS		(uint32_t)0x152000
#define NCS36510_FLASH_CMD_PAGE_ERASE		(uint32_t)0x1
#define NCS36510_FLASH_A_BUSY				(uint32_t)0x1
#define NCS36510_FLASH_B_BUSY				(uint32_t)0x2
#define NCS36510_1K_BYTES					(int)1024
#define NCS36510_PERIPHERAL_FLASH_BIT_POS	(uint32_t)(1 << 23)
#define NCS36510_PERIPHERAL_DEVOPT_BIT_POS	(uint32_t)(1 << 30)
#define NCS36510_FLASH_CONTROL_STALLONWRITE	(uint32_t)0x40
#define NCS36510_FLASH_CONTROL_B_POWERUP	(uint32_t)0x01

#define NCS36510_FLASH_UNLOCK1_CODE			(uint32_t)0xBB781AE9
#define NCS36510_FLASH_UNLOCKAB_CODE		(uint32_t)0xB56D9099
#define NCS36510_FLASH_LOCK1_CODE			(uint32_t)0x12345678
#define NCS36510_FLASH_LOCKAB_CODE			(uint32_t)0x12345678

/* Register memory map */
#define NCS36510_PERIPHERAL_CLOCK			(uint32_t)0x4001B010
#define NCS36510_FLASH_STATUS				(uint32_t)0x40017000
#define NCS36510_FLASH_CONTROL				(uint32_t)0x40017004
#define NCS36510_FLASH_COMMAND				(uint32_t)0x40017008
#define NCS36510_FLASH_ADDRESS				(uint32_t)0x4001700C
#define NCS36510_FLASH_UNLOCK_1				(uint32_t)0x40017010
#define NCS36510_FLASH_UNLOCK_A				(uint32_t)0x40017014
#define NCS36510_WORD_SZ					(uint8_t)4				/* 4 bytes */
#define NCS36510_FLASH_BANK_A				(uint8_t)0
#define NCS36510_FLASH_BANK_B				(uint8_t)1
#define NCS36510_NUM_PAGES_ATOB				(int)512 				/* Gap in pages between flash A and B */
#define NCS36510_DECODE_BUF_SIZE			(int)256

/* Flash page macros */
#define NCS36510_FLASH_PAGE_SIZE			(uint16_t)0x800
#define NCS36510_FLASH_PAGE_MASK			(uint32_t)0xFFFFF800

/* Flash lock/unlock status */
#define NCS36510_FLASH_LOCKED				(uint8_t)1
#define NCS36510_FLASH_UNLOCKED				(uint8_t)0

/* Device option register */
#define NCS36510_DEVICE_OPTION_REG			(uint32_t)0x4001E000
#define NCS36510_DEVICE_OPTION_FULL_FTR		(uint32_t)0x2082353F

/* Default status of flash: Locked */
static uint8_t flash_lock_status = NCS36510_FLASH_LOCKED;

struct ncs36510_flash_bank {
	int probed;
};

struct ncs36510_info {
	uint16_t flash_sz_kib;
	uint16_t ram_sz_kib;
	uint16_t part_num;
	uint16_t part_family;
	uint8_t prod_rev;
	uint16_t page_size;
};

static int ncs36510_write(struct flash_bank *bank, const uint8_t *buffer,
	uint32_t offset, uint32_t count);

static int ncs36510_get_flash_size(struct flash_bank *bank, uint16_t *flash_sz)
{
	*flash_sz = NCS36510_FLASH_SZ;
	return ERROR_OK;
}

static int ncs36510_get_ram_size(struct flash_bank *bank, uint16_t *ram_sz)
{
	*ram_sz = NCS36510_RAM_SZ;
	return ERROR_OK;
}

static int ncs36510_get_part_num(struct flash_bank *bank, uint16_t *pnum)
{
	*pnum = NCS36510_PART_NUM;
	return ERROR_OK;
}

static int ncs36510_get_part_family(struct flash_bank *bank, uint16_t *pfamily)
{
	*pfamily = NCS36510_PART_FAMILY;
	return ERROR_OK;
}

static int ncs36510_get_prod_rev(struct flash_bank *bank, uint8_t *prev)
{
	*prev = NCS36510_PART_REV;
	return ERROR_OK;
}

static int ncs36510_read_info(struct flash_bank *bank,
	struct ncs36510_info *ncs36510_info)
{
	int ret;

	memset(ncs36510_info, 0, sizeof(struct ncs36510_info));

	ret = ncs36510_get_flash_size(bank, &(ncs36510_info->flash_sz_kib));
	if (ERROR_OK != ret)
		return ret;

	ret = ncs36510_get_ram_size(bank, &(ncs36510_info->ram_sz_kib));
	if (ERROR_OK != ret)
		return ret;

	ret = ncs36510_get_part_num(bank, &(ncs36510_info->part_num));
	if (ERROR_OK != ret)
		return ret;

	ret = ncs36510_get_part_family(bank, &(ncs36510_info->part_family));
	if (ERROR_OK != ret)
		return ret;

	ret = ncs36510_get_prod_rev(bank, &(ncs36510_info->prod_rev));
	if (ERROR_OK != ret)
		return ret;

	ncs36510_info->page_size = NCS36510_FLASH_PAGE_SIZE;

	return ERROR_OK;
}

/*
 * Helper to create a human friendly string describing a part
 */

static int ncs36510_decode_info(struct ncs36510_info *info, char *buf,
								 int buf_size)
{
	int printed = 0;

	printed = snprintf(buf, buf_size, "ncs36510 "); /* Part name */

	buf += printed;
	buf_size -= printed;

	if (0 >= buf_size)
	return ERROR_BUF_TOO_SMALL;

	printed = snprintf(buf, buf_size, "Orion"); /* Part family */

	buf += printed;
	buf_size -= printed;

	if (0 >= buf_size)
		return ERROR_BUF_TOO_SMALL;

	/* Product revision */
	printed = snprintf(buf, buf_size, " - Rev: %d", info->prod_rev);
	buf += printed;
	buf_size -= printed;

	if (0 >= buf_size)
		return ERROR_BUF_TOO_SMALL;

	return ERROR_OK;
}

/* flash bank ncs36510 <base> <size> 0 0 <target#>
 */
FLASH_BANK_COMMAND_HANDLER(ncs36510_flash_bank_command)
{
	struct ncs36510_flash_bank *ncs36510_info;

	if (CMD_ARGC < 6)/* TODO 6? */
		return ERROR_COMMAND_SYNTAX_ERROR;

	ncs36510_info = malloc(sizeof(struct ncs36510_flash_bank));

	bank->driver_priv = ncs36510_info;
	ncs36510_info->probed = 0;

	return ERROR_OK;
}

static int ncs36510_erase(struct flash_bank *bank, int first, int last)
{
	/* 1. Enable flash clock
	   2. Unlock flash
	   3. Write FLASH ADDRESS register
	   4. write COMMAND_PAGE_ERASE
	   5. wait until STATUS_NOT_BUSY
	 */

	int ret, page;
	uint32_t clock_config, flash_status;

	/* Read peripheral clock register */
	ret = target_read_u32(bank->target, NCS36510_PERIPHERAL_CLOCK, &clock_config);
	if (ERROR_OK != ret) {
		LOG_ERROR("Failed to read peripheral clock register");
		return ret;
	}

	/* Enable flash clock */
	ret = target_write_u32(bank->target, NCS36510_PERIPHERAL_CLOCK, \
			clock_config & ~NCS36510_PERIPHERAL_FLASH_BIT_POS);
	if (ERROR_OK != ret) {
		LOG_ERROR("Failed to enable flash clock");
		return ret;
	}

	if(bank->bank_number == NCS36510_FLASH_BANK_B) {
		/* Flash B is requested */
		first += NCS36510_NUM_PAGES_ATOB;
		last  += NCS36510_NUM_PAGES_ATOB;
	}

	for(page = first; page <= last; page++)
	{
		/* Unlock flash */
		ret = target_write_u32(bank->target, NCS36510_FLASH_UNLOCK_1, \
							   NCS36510_FLASH_UNLOCK1_CODE);
		if (ERROR_OK != ret) {
			LOG_ERROR("Failed to unlock flash");
			return ret;
		}

		/* Unlock flash A/B */
		ret = target_write_u32(bank->target, (NCS36510_FLASH_UNLOCK_A + \
				(bank->bank_number * NCS36510_WORD_SZ)), \
				 NCS36510_FLASH_UNLOCKAB_CODE);
		if (ERROR_OK != ret) {
			LOG_ERROR("Failed to unlock flash A/B");
			return ret;
		}

		/* Update global lock status */
		flash_lock_status = NCS36510_FLASH_UNLOCKED;

		/* Set address to be erased */
		ret = target_write_u32(bank->target, NCS36510_FLASH_ADDRESS, \
			  (page * NCS36510_FLASH_PAGE_SIZE));
		if (ERROR_OK != ret) {
			LOG_ERROR("Failed to write 'flash address' register");
			return ret;
		}

		/* Execute erase command */
		ret = target_write_u32(bank->target, NCS36510_FLASH_COMMAND, \
			  				NCS36510_FLASH_CMD_PAGE_ERASE);
		if (ERROR_OK != ret) {
			LOG_ERROR("Failed to erase flash");
			return ret;
		}

		/* Check flash erase is complete */
		do
		{
			ret = target_read_u32(bank->target, NCS36510_FLASH_STATUS, \
									&flash_status);
			if (ERROR_OK != ret) {
				LOG_ERROR("Failed to read flash status");
				return ret;
			}
		}while(flash_status & (bank->bank_number + 1)); /* Wait till flash is busy */
	}

	/* Lock flash */
	ret = target_write_u32(bank->target, NCS36510_FLASH_UNLOCK_1, \
						   NCS36510_FLASH_LOCK1_CODE);
	if (ERROR_OK != ret) {
		LOG_ERROR("Failed to lock flash");
		return ret;
	}

	/* Lock flash A/B */
	ret = target_write_u32(bank->target, (NCS36510_FLASH_UNLOCK_A + \
				(bank->bank_number * NCS36510_WORD_SZ)), \
				 NCS36510_FLASH_LOCKAB_CODE);
	if (ERROR_OK != ret) {
		LOG_ERROR("Failed to lock flash A/B");
		return ret;
	}

	/* Update global lock status */
	flash_lock_status = NCS36510_FLASH_LOCKED;

	/* Restore FLASH clock */
	ret = target_write_u32(bank->target, NCS36510_PERIPHERAL_CLOCK, clock_config);
	if (ERROR_OK != ret) {
		LOG_ERROR("Failed to restore flash clock");
		return ret;
	}

	return ERROR_OK;
}

static int ncs36510_protect(struct flash_bank *bank, int set, int first, int last)
{
	int ret;
	uint32_t clock_config;

	if(!set) {
		LOG_ERROR("Erase/write device data to unlock flash");
		return ERROR_FAIL;
	}

	/* Read peripheral clock register */
	ret = target_read_u32(bank->target, NCS36510_PERIPHERAL_CLOCK, &clock_config);
	if (ERROR_OK != ret) {
		LOG_ERROR("Failed to read peripheral clock register");
		return ret;
	}

	/* Enable FLASH clock */
	ret = target_write_u32(bank->target, NCS36510_PERIPHERAL_CLOCK, \
						   clock_config & ~NCS36510_PERIPHERAL_FLASH_BIT_POS);
	if (ERROR_OK != ret) {
		LOG_ERROR("Failed to enable flash clock");
		return ret;
	}

	/* Lock flash */
	ret = target_write_u32(bank->target, NCS36510_FLASH_UNLOCK_1, \
						   NCS36510_FLASH_LOCK1_CODE);
	if (ERROR_OK != ret) {
		LOG_ERROR("Failed to lock flash");
		return ret;
	}

	/* Lock flash A/B */
	ret = target_write_u32(bank->target, (NCS36510_FLASH_UNLOCK_A + \
			(bank->bank_number * NCS36510_WORD_SZ)), NCS36510_FLASH_LOCKAB_CODE);
	if (ERROR_OK != ret) {
		LOG_ERROR("Failed to lock flash A/B");
		return ret;
	}

	/* Update global lock status */
	flash_lock_status = NCS36510_FLASH_LOCKED;

	/* Restore FLASH clock */
	ret = target_write_u32(bank->target, NCS36510_PERIPHERAL_CLOCK, clock_config);
	if (ERROR_OK != ret) {
		LOG_ERROR("Failed to restore flash clock");
		return ret;
	}

	return ERROR_OK;
}

static int ncs36510_write(struct flash_bank *bank, const uint8_t *buffer,
		uint32_t offset, uint32_t count)
{
	int ret;
	uint32_t clock_config, flash_control;

	/* Check flash boundaries */
	if((((offset + count) >= NCS36510_FLASH_A_END_ADDRESS) && \
	   (bank->bank_number == NCS36510_FLASH_BANK_A)) ||
	   ((((offset + count) >= NCS36510_FLASH_B_END_ADDRESS) || \
	   (offset < NCS36510_FLASH_B_START_ADDRESS)) && \
	   (bank->bank_number == NCS36510_FLASH_BANK_B))) {
		/* Outside flash boundary */
		LOG_ERROR("Trying to write outside the flash for bank number %d at address 0x%08X", \
					bank->bank_number, (offset + count));
		return ERROR_FAIL;
	}

	/* Read peripheral clock register */
	ret = target_read_u32(bank->target, NCS36510_PERIPHERAL_CLOCK, &clock_config);
	if (ERROR_OK != ret) {
		LOG_ERROR("Failed to read peripheral clock register");
		return ret;
	}

	/* Enable FLASH clock */
	ret = target_write_u32(bank->target, NCS36510_PERIPHERAL_CLOCK, \
					clock_config & ~NCS36510_PERIPHERAL_FLASH_BIT_POS);
	if (ERROR_OK != ret) {
		LOG_ERROR("Failed to enable flash clock");
		return ret;
	}

	/* Unlock flash */
	ret = target_write_u32(bank->target, NCS36510_FLASH_UNLOCK_1, \
				NCS36510_FLASH_UNLOCK1_CODE);
	if (ERROR_OK != ret) {
		LOG_ERROR("Failed to unlock flash");
		return ret;
	}

	/* Unlock flash A/B */
	ret = target_write_u32(bank->target, (NCS36510_FLASH_UNLOCK_A + \
				(bank->bank_number * NCS36510_WORD_SZ)), \
				 NCS36510_FLASH_UNLOCKAB_CODE);
	if (ERROR_OK != ret) {
		LOG_ERROR("Failed to unlock flash A/B");
		return ret;
	}

	/* Update global lock status */
	flash_lock_status = NCS36510_FLASH_UNLOCKED;

	/* Read flash control register */
	ret = target_read_u32(bank->target, NCS36510_FLASH_CONTROL, &flash_control);
	if (ERROR_OK != ret) {
		LOG_ERROR("Failed to read flash control register");
		return ret;
	}

	/* Make sure the CM3 hangs while write is ongoing   */
	ret = target_write_u32(bank->target, NCS36510_FLASH_CONTROL, \
					flash_control | NCS36510_FLASH_CONTROL_STALLONWRITE);
	if (ERROR_OK != ret) {
		LOG_ERROR("Failed to stall on write");
		return ret;
	}

	/* Write code into flash */
	uint32_t index;
	for(index = 0; index < (count / NCS36510_WORD_SZ); index++)
	{
		/* Write multiple of words */
		ret = target_write_u32(bank->target, (offset + index * NCS36510_WORD_SZ), \
				  *((uint32_t*)(buffer + (index * NCS36510_WORD_SZ))));
		if (ERROR_OK != ret) {
			LOG_ERROR("Failed to write to flash");
			return ret;
		}
	}

	if((count % NCS36510_WORD_SZ) > 0)
	{
		/* Write remaining bytes */
		ret = target_write_buffer(bank->target, (offset + (index * NCS36510_WORD_SZ)), \
				(count % NCS36510_WORD_SZ), (buffer + (index * NCS36510_WORD_SZ)));
		if (ERROR_OK != ret) {
			LOG_ERROR("Failed to write to flash");
			return ret;
		}
	}

	/* Lock flash */
	ret = target_write_u32(bank->target, NCS36510_FLASH_UNLOCK_1, \
						   NCS36510_FLASH_LOCK1_CODE);
	if (ERROR_OK != ret) {
		LOG_ERROR("Failed to lock flash");
		return ret;
	}

	/* Lock flash A/B */
	ret = target_write_u32(bank->target, (NCS36510_FLASH_UNLOCK_A + \
			(bank->bank_number * NCS36510_WORD_SZ)), NCS36510_FLASH_LOCKAB_CODE);
	if (ERROR_OK != ret) {
		LOG_ERROR("Failed to lock flash A/B");
		return ret;
	}

	/* Update global lock status */
	flash_lock_status = NCS36510_FLASH_LOCKED;

	/* Restore flash clock */
	ret = target_write_u32(bank->target, NCS36510_PERIPHERAL_CLOCK, clock_config);
	if (ERROR_OK != ret) {
		LOG_ERROR("Failed to restore flash clock");
		return ret;
	}

	return ERROR_OK;
}

static int ncs36510_probe(struct flash_bank *bank)
{
	struct ncs36510_flash_bank *ncs36510_info = bank->driver_priv;
	struct ncs36510_info ncs36510_mcu_info;
	int ret, i;
	int num_pages;
	uint32_t base_address, clock_config, flash_control, flash_status;
	char buf[NCS36510_DECODE_BUF_SIZE];

	ncs36510_info->probed = 0;

	/* Read peripheral clock register */
	ret = target_read_u32(bank->target, NCS36510_PERIPHERAL_CLOCK, &clock_config);
	if (ERROR_OK != ret) {
		LOG_ERROR("Failed to read peripheral clock register");
		return ret;
	}

	/* Enable FLASH clock */
	ret = target_write_u32(bank->target, NCS36510_PERIPHERAL_CLOCK, clock_config & \
	    ~(NCS36510_PERIPHERAL_FLASH_BIT_POS | NCS36510_PERIPHERAL_DEVOPT_BIT_POS));
	if (ERROR_OK != ret) {
		LOG_ERROR("Failed to enable flash clock");
		return ret;
	}

	/* Enable full featured device(Both flash A and B enabled) if not */
	ret = target_write_u32(bank->target, NCS36510_DEVICE_OPTION_REG, \
					NCS36510_DEVICE_OPTION_FULL_FTR);
	if (ERROR_OK != ret) {
		LOG_ERROR("Failed to enable full featured device");
		return ret;
	}

	/* Read flash control register */
	ret = target_read_u32(bank->target, NCS36510_FLASH_CONTROL, &flash_control);
	if (ERROR_OK != ret) {
		LOG_ERROR("Failed to read flash control register");
		return ret;
	}

	/* Power up flash B  */
	ret = target_write_u32(bank->target, NCS36510_FLASH_CONTROL, \
					flash_control & ~NCS36510_FLASH_CONTROL_B_POWERUP);
	if (ERROR_OK != ret) {
		LOG_ERROR("Failed to power up flash B");
		return ret;
	}

	do
	{
		ret = target_read_u32(bank->target, NCS36510_FLASH_STATUS, &flash_status);
		if (ERROR_OK != ret) {
			LOG_ERROR("Failed to read flash status");
			return ret;
		}
	}while(flash_status & NCS36510_FLASH_B_BUSY); /* Wait till flash is busy */

	/* Restore flash clock */
	ret = target_write_u32(bank->target, NCS36510_PERIPHERAL_CLOCK, clock_config);
	if (ERROR_OK != ret) {
		LOG_ERROR("Failed to restore flash clock");
		return ret;
	}

	if(bank->bank_number == NCS36510_FLASH_BANK_A)
	{
		base_address = NCS36510_FLASH_A_START_ADDRESS;
	}
	else if(bank->bank_number == NCS36510_FLASH_BANK_B)
	{
		base_address = NCS36510_FLASH_B_START_ADDRESS;
	}
	else
	{
		/* Invalid bank number('else' added to satisfy the compiler) */
		LOG_ERROR("Invalid bank number %x", bank->bank_number);
		return ERROR_FAIL;
	}

	ret = ncs36510_read_info(bank, &ncs36510_mcu_info);
	if (ERROR_OK != ret)
		return ret;


	ret = ncs36510_decode_info(&ncs36510_mcu_info, buf, sizeof(buf));
	if (ERROR_OK != ret)
		return ret;

	LOG_INFO("detected part: %s", buf);
	LOG_INFO("flash size = %dkbytes", ncs36510_mcu_info.flash_sz_kib);
	LOG_INFO("flash page size = %dbytes", ncs36510_mcu_info.page_size);

	assert(0 != ncs36510_mcu_info.page_size);

	num_pages = ncs36510_mcu_info.flash_sz_kib * NCS36510_1K_BYTES / \
				ncs36510_mcu_info.page_size;

	assert(num_pages > 0);

	if (bank->sectors) {
		free(bank->sectors);
		bank->sectors = NULL;
	}

	bank->base = base_address;
	bank->size = (num_pages * ncs36510_mcu_info.page_size);
	bank->num_sectors = num_pages;

	bank->sectors = malloc(sizeof(struct flash_sector) * num_pages);

	for (i = 0; i < num_pages; i++) {
		bank->sectors[i].offset = i * ncs36510_mcu_info.page_size;
		bank->sectors[i].size = ncs36510_mcu_info.page_size;
		bank->sectors[i].is_erased = -1;
		bank->sectors[i].is_protected = NCS36510_FLASH_LOCKED;
	}

	ncs36510_info->probed = 1;

	return ERROR_OK;
}

static int ncs36510_auto_probe(struct flash_bank *bank)
{
	struct ncs36510_flash_bank *ncs36510_info = bank->driver_priv;

	if (ncs36510_info->probed)
		return ERROR_OK;

	return ncs36510_probe(bank);
}

static int ncs36510_protect_check(struct flash_bank *bank)
{
	int i;

	/* Update global lock status */
	for (i = 0; i < bank->num_sectors; i++) {
		bank->sectors[i].is_protected = flash_lock_status;
	}

	return ERROR_OK;
}

static int ncs36510_get_info(struct flash_bank *bank, char *buf, int buf_size)
{
	struct ncs36510_info info;
	int ret = 0;

	ret = ncs36510_read_info(bank, &info);
	if (ERROR_OK != ret) {
		LOG_ERROR("Failed to read NCS36510 info");
		return ret;
	}

	return ncs36510_decode_info(&info, buf, buf_size);
}

COMMAND_HANDLER(ncs36510_handle_debuglock_command)
{
	return ERROR_OK;
}

static const struct command_registration ncs36510_exec_command_handlers[] = {
	{
		.name = "debuglock",
		.handler = ncs36510_handle_debuglock_command,
		.mode = COMMAND_EXEC,
		.usage = "bank_id",
		.help = "Lock the debug interface of the device.",
	},
	COMMAND_REGISTRATION_DONE
};

static const struct command_registration ncs36510_command_handlers[] = {
	{
		.name = "ncs36510",
		.mode = COMMAND_ANY,
		.help = "ncs36510 flash command group",
		.usage = "",
		.chain = ncs36510_exec_command_handlers,
	},
	COMMAND_REGISTRATION_DONE
};

struct flash_driver ncs36510_flash = {
	.name = "ncs36510",
	.commands = ncs36510_command_handlers,
	.flash_bank_command = ncs36510_flash_bank_command,
	.erase = ncs36510_erase,
	.protect = ncs36510_protect,
	.write = ncs36510_write,
	.read = default_flash_read,
	.probe = ncs36510_probe,
	.auto_probe = ncs36510_auto_probe,
	.erase_check = default_flash_blank_check,
	.protect_check = ncs36510_protect_check,
	.info = ncs36510_get_info,
};